#!/usr/bin/env python

from __future__ import with_statement

import sys
import os
import socket
import signal
from StringIO import StringIO
import thread
import time
import UserDict

from vmailmgr.decoder import Decoder
from vmailmgr.encoder import Encoder

from mail.send import SendMail


class Error (Exception):
	pass

CONNECTION_TIMEOUT=1

def send_alert (entries):
	try:
		mail = SendMail()
		mail.subject('vmailmgr proxy issue')
		mail.sender('support@exa-networks.co.uk')
		mail.recipient('support@exa-networks.co.uk')
		mail.send('There was a problem when trying to find destinations for handled domains. The problem entries are:\n\n%s' % str(entries))
	except:
		pass


class Backend:
	def __init__(self,timeout=30):
		self.tracker = {}
		self.timeout = timeout
		self.lock = thread.allocate_lock()

	def reset(self):
		with self.lock:
			self.tracker = {}

	def mark(self, ip):
		with self.lock:
			self.tracker[ip] = time.time()

	def wait(self, ip):
		with self.lock:
			now = time.time()
			last_error = self.tracker.get(ip, 0)
			if (now - last_error) < self.timeout:
				return True
		return False

backend = Backend()

class KeyValue (UserDict.UserDict):
	def __init__(self,folder,file):
		UserDict.UserDict.__init__(self)
		self.line = 0
		self.file = file
		self.fname = ""
		self.errors = []

		for entry in self.parseFile(os.path.join(folder,file)):
			try:
				key, value = entry
			except ValueError, e:
				self.errors.append(entry)

			key = key.lower()
			value = value.lower()
			self[key] = value


	def parseFile (self,name):
		self.number = 0
		self.fname = name
		n = os.path.split(name)[1]
		file = open(name,'r')
		for self.line in file.readlines():
			self.number += 1
			l = self.line.strip()
			if len(l) == 0:
				continue
			if l[0] == '#':
				continue
			yield self.parseLine(self.number,l)
		file.close()

	def parseLine (self,number,string):
		result = []
		parts = string.split(':')
		for part in parts:
			part = part.replace('\\072',':')
			result.append(part)
		return result

def log (msg):
	print >> sys.stderr, msg

pid = os.getpid()
domain_conf = {}
config_lock = thread.allocate_lock()

def reload():
	global domain_conf
	global config_lock
	global backend

	log('reloading proxy configuration')
	# Loading configuration
	try:
		with config_lock:
			signal.signal(signal.SIGHUP, handler_hup)
			backend.reset()

			domain_conf = {}
			domains = KeyValue('destination','domain.conf')
			servers = KeyValue('destination','server.conf')

			if domains.errors:
				send_alert(domain.errors)
			if servers.errors:
				send_alert(domain.errors)

			for domain,destination in domains.iteritems():
				if servers.has_key(destination):
					ip = servers[destination]
					log('configuration: ip %s for domain %s' % (ip, domain))
					domain_conf[domain] = ip
				else:
					log('configuration: could not find a valid server (%s) for %s' % (destination,domain))
					continue

	except Error:
		log('proxy failed to read configuration')
		sys.exit(1)

def handler_hup (signum,frame):
	reload()

# svc -d first signal
def handler_cont (signum,frame):
	log('exiting ...')
	global pid
	os.kill(pid,9)

# svc -d second signal
def handler_term (signum,frame):
	log('exiting ...')
	global pid
	os.kill(pid,9)

signal.signal(signal.SIGCONT, handler_cont)
signal.signal(signal.SIGTERM, handler_term)

reload()

class Fetch (object):
	def __init__(self,ip=None,port=322):
		self.ip = ip
		self.port = port
		self.csocket = None
	
	def __del__(self):
		self.close()

	def write (self,string):
		return self.csocket.send(string)
	
	def read (self,n):
		return self.csocket.recv(n)
	
	def connect (self,ip=None,port=None):
		if ip != None:
			self.ip = ip
		
		if port != None:
			self.port = port
				
		if self.ip != None and self.port != None:
			self.close()
			self.csocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
			self.csocket.settimeout(CONNECTION_TIMEOUT)
			self.csocket.connect((self.ip, self.port))
	
	def close (self):
		if self.csocket != None:
			try: self.csocket.close()
			except: pass
			self.csocket = None

class Proxy (object):
	def __init__ (self,recv=sys.stdin,send=sys.stdout):
		self.error = None
		self.fclient = StringIO(recv)
		self.tclient = StringIO(send)
		self.encoder = Encoder()
		self.decoder = Decoder(self.fclient)
		self.debug = True

	def failed (self,message):
		try:
			self.error = message
			self.tclient.write(self.encoder.encode_response(3,self.error))
		except:
			self.error += ' and could not notify client of it'

		log(self.error)

	def process (self):
		try:
			arg = self.decoder.decode_message()
			# Security issue if left
			# log('parsed received data is %s' % str(arg))
		except:
			self.failed('Could not decode request sent')
			return False
	
		global config_lock
		with config_lock:
			try:
				domain = arg[1].lower()
				ip = domain_conf[domain]
			except:
				self.failed('proxy could not find backend for %s' % domain)
				return False
		
		if not backend.wait(ip):
			try:
				fetch = Fetch()
				fetch.connect(ip)
				fetch.write(self.decoder.seen)
			except socket.error:
				self.failed('problem connecting to backend server at %s' % arg[1])
				fetch.close()
				backend.mark(ip)
				return False
		else:
			self.failed('recent error on %s - waiting' % ip)
			return False
		
		try:
			decoder = Decoder(fetch)
			code, response = decoder.decode_response()
		except:
			self.failed('problem parsing data received from backend')
			fetch.close()
			return False
		
		#if code == 0:
		#	log('backend served request')
		#else:
		#	log('backend said, ('+ str(code) + ') reponse: ' + str(response))
		
		try:
			self.tclient.write(decoder.seen)
		except socket.error:
			log('problem responding to the client')
			fetch.close()
			return False

		fetch.close()
		return True
	
from twisted.application import internet, service
from twisted.internet import protocol, reactor
from twisted.protocols import basic

class ProxyProtocol(basic.LineReceiver):
	def dataReceived (self,data):
		proxy = Proxy(data,'')
		
		try:
			proxy.process()
		except:
			log('unknown problem')
			self.transport.loseConnection()
			# XXX: Left for debugging, if it works, remove
			raise
			return
		
		self.transport.write(proxy.tclient.getvalue())
		self.transport.loseConnection()

class ProxyFactory(protocol.ServerFactory):
	protocol = ProxyProtocol

application = service.Application('proxy')
factory = ProxyFactory()
internet.TCPServer(322, factory).setServiceParent(service.IServiceCollection(application))

